import rbo
import numpy as np
import swifter
from scipy import stats
from typing import List

from .normalise import normalise_for_kendalltau


def ndcg(true_relevance: List[float],
         predicted_relevance: List[float],
         cutoff: int) -> float:
    """
    :param true_relevance: List of scores corresponding to the ground truth items.
    :param predicted_relevance: List of scores corresponding to the predicted items.
    :param cutoff: The depth of the list to consider for rank validation.
    :return:
    """

    def cumm_gain(relevance_scores: List[float], log_vals: List[int], cutoff: int):
        """
        :param relevance_scores: List of scores corresponding to a ranked list of items.
        :param log_vals: log values
        :param cutoff: The depth of the list to consider for rank validation.
        :return:
        """
        cumm_sum = 0
        for index, relevance_value in enumerate(relevance_scores[:cutoff]):
            num = 2 ** relevance_value - 1
            den = log_vals[index]
            cumm_sum += num / den
        return cumm_sum

    log_vals = np.log2([index + 1 for index in range(1, cutoff + 1)])
    dcg = cumm_gain(predicted_relevance, log_vals, cutoff)
    ideal_dcg = cumm_gain(true_relevance, log_vals, cutoff)
    ndcg_score = dcg / ideal_dcg
    return ndcg_score


def recall(true_relevance: List[str],
           predicted_relevance: List[str],
           cutoff: int) -> float:
    """
    :param true_relevance: List of ground truth items.
    :param predicted_relevance: List of predicted items.
    :param cutoff: The depth of the list to consider for rank validation.
    :return: An integer score
    """

    intersection = set(true_relevance[:cutoff]) & set(predicted_relevance[:cutoff])
    recall_score = len(intersection) / len(true_relevance)
    return recall_score


def kendall_tau(true_relevance:  List[str],
                predicted_relevance:  List[str],
                cutoff: int) -> float:
    """
    :param true_relevance: List of ground truth items.
    :param predicted_relevance: List of predicted items.
    :param cutoff: The depth of the list to consider for rank validation.
    :return: An integer score
    """

    true_relevance, predicted_relevance = normalise_for_kendalltau(true_relevance, predicted_relevance, cutoff)
    tau, _ = stats.kendalltau(true_relevance[:cutoff], predicted_relevance[:cutoff])
    return tau


def rbo_sim(true_relevance:  List[str],
            predicted_relevance:  List[str],
            cutoff: int) -> float:
    """
    :param true_relevance: List of ground truth items.
    :param predicted_relevance: List of predicted items.
    :param cutoff: The depth of the list to consider for rank validation.
    :return: An integer score
    """

    rbo_score = rbo.RankingSimilarity(true_relevance[:cutoff], predicted_relevance[:cutoff]).rbo()
    return rbo_score



import collections
from typing import List, Tuple


class OrderedSet(collections.Set):
    def __init__(self, iterable=()):
        self.d = collections.OrderedDict.fromkeys(iterable)

    def __len__(self):
        return len(self.d)

    def __contains__(self, item):
        return item in self.d

    def __iter__(self):
        return iter(self.d)


def normalise_for_kendalltau(true_relevance: List[str],
                             predicted_relevance: List[str],
                             cutoff: int) -> Tuple[List[str], List[str]]:
    """

    :param true_relevance: list of ground truth items (doc ids).
    :param predicted_relevance: list of predicted items (doc ids).
    :param cutoff: The depth of the list to consider for rank validation.
    :return: Normalised true relevance items (doc ids) and normalised predicted relevance items (doc ids)
    """

    add_to_true_rel = list(OrderedSet(predicted_relevance[:cutoff]) - OrderedSet(true_relevance[:cutoff]))
    add_to_pred_rel = list(OrderedSet(true_relevance[:cutoff]) - OrderedSet(predicted_relevance[:cutoff]))
    norm_true_rel = list(OrderedSet(true_relevance[:cutoff])) + add_to_true_rel
    norm_pred_rel = list(OrderedSet(predicted_relevance[:cutoff])) + add_to_pred_rel
    return norm_true_rel, norm_pred_rel


def normalise_relevance(true_relevance_items: List[str],
                        true_relevance_scores: List[float],
                        predicted_relevance_items: List[str]) -> Tuple[List[Tuple[str, float]], List[Tuple[str, float]]]:
    """

    :param true_relevance_items: list of ground truth items (doc ids)
    :param true_relevance_scores: list of relevancy scores, corresponding to the items(doc ids) in truth_items
    :param predicted_relevance_items: list of predicted lists of the items(doc ids)
    :return: Two list of tuples with (item, score), each for ground truth and the prediction
    """

    normalised_true_relevance = list(zip(true_relevance_items, true_relevance_scores))
    true_relevance_dict = dict(normalised_true_relevance)
    normalised_predicted_relevance = []  # list of tuples
    for predicted_item in predicted_relevance_items:
        if predicted_item in true_relevance_dict:
            normalised_predicted_relevance.append((predicted_item, true_relevance_dict[predicted_item]))
        else:
            normalised_predicted_relevance.append((predicted_item, 0))
    return normalised_true_relevance, normalised_predicted_relevance



import swifter

import pandas as pd
from typing import List, Tuple

from .metrics import ndcg, recall, kendall_tau, rbo_sim
from .normalise import normalise_relevance


def prepare_relevance(truth_items: List[str],
                      truth_scores: List[float],
                      pred_items: List[str]) -> pd.Series:
    """
    :param truth_items:  list of ground truth items (doc ids)
    :param truth_scores: list of relevancy scores, corresponding to the items(doc ids) in truth_items
    :param pred_items:   list of predicted lists of the items(doc ids)
    :return: normalised and prepared ground truth items, scores corresponding to the ground truth items(doc ids), predicted list of ranked items,
             scores of items(doc ids) corresponding to the predicted list of items(doc ids).
    """
    truth_rel, pred_rel = normalise_relevance(truth_items, truth_scores, pred_items)
    truth_rel_items, truth_rel_scores = list(map(list, zip(*truth_rel)))
    pred_rel_items, pred_rel_scores = list(map(list, zip(*pred_rel)))
    return pd.Series([truth_rel_items, truth_rel_scores, pred_rel_items, pred_rel_scores])


def create_metric_cols(df: pd.DataFrame,
                       truth_item_col: pd.Series,
                       truth_score_col: pd.Series,
                       pred_item_col: pd.Series,
                       pred_score_col: pd.Series,
                       metric_list: List[str],
                       cutoff_list: List[int]) -> pd.DataFrame:
    """
    :param df: A dataframe with columns : Query, truth_item_col, truth_score_col, pred_item_col, pred_score_col
    :param truth_item_col: Dataframe column of ground truth items(doc ids) (Each entry of the column should be a list)
    :param truth_score_col: Dataframe column of ground truth scores (Each entry of the column should be a list)
    :param pred_item_col: Dataframe column of predicted items(doc ids) (Each entry of the column should be a list)
    :param pred_score_col: Dataframe column of predicted scores (Each entry of the column should be a list)
    :param metric_list: List of metrics to be chose from ['ndcg', 'kendall_tau', 'rbo', 'recall']
    :param cutoff_list: List of different cutoffs (corresponding to the depth of the list) for which we want the validation metrics.
    :return: A dataframe with the desired metric report
    """
    if "ndcg" in metric_list:
        for cutoff in cutoff_list:
            col_name = "ndcg@" + str(cutoff)
            df[col_name] = df.swifter.apply(lambda x: ndcg(x[truth_score_col], x[pred_score_col], cutoff), axis=1)

    if "recall" in metric_list:
        for cutoff in cutoff_list:
            col_name = "recall@" + str(cutoff)
            df[col_name] = df.swifter.apply(lambda x: recall(x[truth_item_col], x[pred_item_col], cutoff), axis=1)

    if "kendall_tau" in metric_list:
        for cutoff in cutoff_list:
            col_name = "kendall_tau@" + str(cutoff)
            df[col_name] = df.swifter.apply(lambda x: kendall_tau(x[truth_item_col], x[pred_item_col], cutoff), axis=1)

    if "rbo" in metric_list:
        for cutoff in cutoff_list:
            col_name = "rbo@" + str(cutoff)
            df[col_name] = df.swifter.apply(lambda x: rbo_sim(x[truth_item_col], x[pred_item_col], cutoff), axis=1)

    return df


def get_metrics_report(df: pd.DataFrame,
                       truth_item_col: pd.Series,
                       truth_score_col: pd.Series,
                       pred_item_col: pd.Series,
                       metric_list: List[str],
                       cutoff_list: List[int]) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    :param df: A dataframe with columns : Query, truth_item_col, truth_score_col, pred_item_col, pred_score_col.
    :param truth_item_col: A dataframe column with ground truth list of items(doc ids).
    :param truth_score_col: A dataframe column with list of scores, corresponding to the items(doc ids) in ground truth col
    :param pred_item_col: A dataframe column with predicted ranked lists.
    :param metric_list: List of metrics to be chosen from ['ndcg', 'kendall_tau', 'rbo', 'recall'].
    :param cutoff_list: List of different cutoffs (corresponding to the depth of the list) for which we want the validation metrics.
    :return: Two dataframes with query level report and overall summary report.
    """

    df[[truth_item_col, truth_score_col, pred_item_col, "pred_score_col"]] = df.swifter.apply(
        lambda x: prepare_relevance(x[truth_item_col],
                                    x[truth_score_col],
                                    x[pred_item_col]),
        axis=1)
    df_scored = create_metric_cols(df, truth_item_col, truth_score_col, pred_item_col, "pred_score_col",
                                   metric_list, cutoff_list)
    query_level_report = df_scored.drop([truth_item_col, truth_score_col, pred_item_col, "pred_score_col"], axis=1)
    overall_summary_report = query_level_report.describe()
    return query_level_report, overall_summary_report



from setuptools import setup, find_packages

setup(
    name="rank_validation",
    version="1.0.1",
    description="creates a rank validation report",
    py_modules=["validation_generator", "metrics", "normalise"],
    install_requires=["numpy", "pandas", "swifter", "rbo", "scipy", "typing"],
    author="Akash Dubey",
    url="https://github.com/akashkdubey/ranking_validation",
    author_email="akashdubey826@gmail.com",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Information Analysis"
    ]
)
